/* ####################################################
   # Copyright 2002, 2003 Ren Serral,                #
   #                                                  #
   # sequence.c - Flow Management.                    #
   #       v2.0 (04-2002)                             #
   #       v2.01(02-2003)                             #
   ####################################################

   This file is part of NetMeter.

   NetMeter is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   NetMeter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with NetMeter; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <stdio.h>

#include "sequence.h"
#include "common.h"

/* Constructs a list of the lost packets of a given flow */
p_slist_head new_slist(p_flist_head item, int flow_id) {
	p_slist_head list;
	if (!item) {
#ifdef DEBUG
		fprintf(stderr,"Trying to create an empty sequence list\n");
#endif
		return NULL;
	}

	list = (p_slist_head)malloc(sizeof(slist_head));
	list->the_flow = copy_flow(item, flow_id);
	list->flow_id = flow_id;
	list->head = 0;
	list->tail = 0;
	list->elements = 0;

	if (do_holes(list) < 0) {
		free(list);
		return NULL;
	}

	return list;
}

/* Returns a pointer to the first hole of the list */
p_hole go_sfirst(p_slist_head list) {
	if (!list) {
#ifdef DEBUG
		fprintf(stderr,"Can't go first item of a NULL list!\n");
#endif
		return NULL;
	}
	list->actual = list->head;
	return list->actual;
}

p_hole go_slast(p_slist_head list) {
	if (!list) {
#ifdef DEBUG
		fprintf(stderr,"Can't go first item of a NULL list!\n");
#endif
		return NULL;
	}
	list->actual = list->tail;
	return list->actual;
}

p_hole go_snext(p_slist_head list) {
	if (!list) {
#ifdef DEBUG
		fprintf(stderr,"Can't go first item of a NULL list!\n");
#endif
		return NULL;
	}
	if (!list->actual)
		list->actual = list->head;
	else
		list->actual = list->actual->next;
	if (!list->actual) {
		list->actual = list->tail;
		return NULL;
	}
	return list->actual;
}

p_hole go_sprev(p_slist_head list) {
	if (!list) {
#ifdef DEBUG
		fprintf(stderr,"Can't go first item of a NULL list!\n");
#endif
		return NULL;
	}
	if (!list->actual)
		list->actual = list->tail;
	else
		list->actual = list->actual->prev;
	if (!list->actual) {
		list->actual = list->head;
		return NULL;
	}
	return list->actual;
}

int del_slist(p_slist_head list) {
	p_hole tmp, next;

	if (!list)
		return -MYEINVAL;
	tmp = list->head;
	while (tmp != list->tail) {
		next = tmp->next;
		free(tmp);
		tmp = next;
	}
	free(tmp);
	del_list(list->the_flow);
	free(list);
	return OK;
}

int hole_duration(p_hole myHole, struct timeval *myTime) {
	if (!myHole || !myTime) {
		return -MYEINVAL;
	}
	return time_diff(&(myHole->last->rx_time), &(myHole->first->rx_time), myTime);
}

/* Static members (private) */
int do_holes(p_slist_head list) {
	char *sequence;
	int flow_id;
	int tmp_seq;
	p_flow tmp;

	if (!list) {
#ifdef DEBUG
		fprintf(stderr, "No list to look at\n");
#endif
		return -MYEINVAL;
	}
	flow_id = list->flow_id;
	sequence = list->sequence;
	bzero(sequence, sizeof(list->sequence));

	tmp = go_first(list->the_flow);
	if (!tmp)
		return -MYEINVAL;
	do {
		if (tmp->flow_id == flow_id) {
			tmp_seq = tmp->seq_num >> 3;
			if ((sequence[tmp_seq] & MASK(tmp->seq_num % 8))) {
#ifdef DEBUG
				fprintf(stderr, "Duplicated sequence number %d for flow: %d\n", tmp->seq_num,tmp->flow_id);
#endif
				return -MYEINVAL;
			}
			sequence[tmp_seq] = sequence[tmp_seq] | MASK(tmp->seq_num % 8);
		}
	} while (tmp = go_next(list->the_flow));
	if (make_hole_list(list) < 0)
		return -MYEINVAL;
	return OK;
}

int make_hole_list(p_slist_head list) {
	int i, j, tmp_seq;
	char *sequence;
	if (!list) {
#ifdef DEBUG
		fprintf(stderr, "No list to look at\n");
#endif
		return -MYEINVAL;
	}
	sequence = list->sequence;
	i = 0;
	while (i < list->the_flow->max_seq) {
		j = i;
		while (!(sequence[i>>3] & MASK(i % 8)) && (i < list->the_flow->max_seq))
			i++;
		if (j != i)
			add_hole(list, j, i - 1);
		while ((sequence[i>>3] & MASK(i % 8)) && (i < list->the_flow->max_seq))
			i++;
	}
	return OK;
}

int add_hole(p_slist_head list, int begin_, int end_) {
	int begin, end, i;
	p_flow tmp_begin, tmp_end;
	p_hole new_item;

	if (!list) {
#ifdef DEBUG
		fprintf(stderr, "No list to look at\n");
#endif
		return -MYEINVAL;
	}
	if (end_ < begin_) {
		begin = end_;
		end = begin_;
	}
	else {
		begin = begin_;
		end = end_;
	}
	tmp_begin = go_first(list->the_flow);
	while (tmp_begin && tmp_begin->seq_num != (begin - 1))
		tmp_begin = go_next(list->the_flow);
	if (!tmp_begin) {
		return -MYEINVAL;
	}
	if (begin == end)
		tmp_end = tmp_begin;
	else {
		tmp_end = go_first(list->the_flow);
		while (tmp_end && tmp_end->seq_num != (end + 1))
			tmp_end = go_next(list->the_flow);
		if (!tmp_end) {
			return -MYEINVAL;
		}
	}
	new_item = (p_hole)malloc(sizeof(hole));
	if (!new_item) {
#ifdef DEBUG
		fprintf(stderr,"I don't have enought memory, sorry!\n");
#endif
		return -MYEINVAL;
	}
	new_item->first = tmp_begin;
	new_item->last = tmp_end;
	new_item->packets = end - begin + 1;
	new_item->next = 0;
	if (list->tail) {
		new_item->prev = list->tail;
		list->tail->next = new_item;
	} else {	// New List
		new_item->prev = 0;
		list->head = new_item;
	}
	list->tail = new_item;
	list->elements++;
	return OK;

}
